package net.sf.doolin.util;

import java.awt.Color;
import java.awt.color.ColorSpace;

/**
 * This class allows to compute a gradient between two colors, according to a
 * given color space.
 * 
 * @author Damien Coraboeuf
 * 
 * 
 */
public class Gradient {

	private final ColorSpace colorSpace;
	private final Color ca;
	private final Color cb;

	/**
	 * Creates a gradient between two colors using the <code>sRGB</code> color
	 * space.
	 * 
	 * @param ca
	 *            First color
	 * @param cb
	 *            Second color
	 */
	public Gradient(Color ca, Color cb) {
		this(ca, cb, ColorSpace.getInstance(ColorSpace.CS_sRGB));
	}

	/**
	 * Creates a gradient between two colors using the given color space.
	 * 
	 * @param ca
	 *            First color
	 * @param cb
	 *            Second color
	 * @param colorSpace
	 *            Color space used for the computation of the gradient
	 */
	public Gradient(Color ca, Color cb, ColorSpace colorSpace) {
		this.ca = ca;
		this.cb = cb;
		this.colorSpace = colorSpace;
	}

	/**
	 * Gets the first color
	 * 
	 * @return Color
	 */
	public Color getCa() {
		return this.ca;
	}

	/**
	 * Gets the second color
	 * 
	 * @return Color
	 */
	public Color getCb() {
		return this.cb;
	}

	/**
	 * Computes the color between the two ends, using the given percentage.
	 * 
	 * @param percent
	 *            Percentage between the first color (0% - 0.0) and the second
	 *            color (100% - 1.0). Any value lesser than 0.0 is changed to 0,
	 *            and any value greater than is changed to 1.
	 * @return Intermediate color
	 * @see Color#getColorComponents(ColorSpace, float[])
	 */
	public Color getColor(double percent) {
		if (percent < 0) {
			percent = 0;
		} else if (percent > 1) {
			percent = 1;
		}
		// Colors components
		int n = this.colorSpace.getNumComponents();
		float[] fa = new float[n];
		float[] fb = new float[n];
		this.ca.getColorComponents(this.colorSpace, fa);
		this.cb.getColorComponents(this.colorSpace, fb);
		// Computation
		float[] x = new float[n];
		for (int i = 0; i < n; i++) {
			x[i] = (float) (fa[i] + (fb[i] - fa[i]) * percent);
		}
		// Conversion
		x = this.colorSpace.toRGB(x);
		Color color = new Color(x[0], x[1], x[2]);
		// OK
		return color;
	}

	/**
	 * Gets the color space used by this gradient.
	 * 
	 * @return Color space
	 */
	public ColorSpace getColorSpace() {
		return this.colorSpace;
	}

}
