/*
 * Created on 10 mars 2005
 */
package net.sf.doolin.util;

import com.google.common.base.Function;
import com.google.common.collect.Collections2;
import com.google.common.collect.Lists;
import net.sf.jstring.LocalizableException;
import org.apache.commons.beanutils.BeanUtils;
import org.apache.commons.beanutils.MethodUtils;
import org.apache.commons.beanutils.PropertyUtils;
import org.apache.commons.lang.ObjectUtils;
import org.apache.commons.lang.SerializationUtils;
import org.apache.commons.lang.StringUtils;
import org.apache.commons.lang.Validate;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import javax.swing.*;
import java.awt.*;
import java.beans.PropertyDescriptor;
import java.io.*;
import java.lang.annotation.Annotation;
import java.lang.reflect.Constructor;
import java.lang.reflect.Field;
import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;
import java.net.URL;
import java.text.ParseException;
import java.text.SimpleDateFormat;
import java.util.*;
import java.util.List;
import java.util.regex.Pattern;

/**
 * Set of utility methods.
 * 
 * @author Damien Coraboeuf
 */
public class Utils {
	/**
	 * Logger
	 */
	private final static Logger log = LoggerFactory.getLogger(Utils.class);

	/**
	 * ISO format
	 */
	public static final String ISO_DATE_FORMAT = "yyyy-MM-dd'T'HH:mm:ss.SSS";

	/**
	 * Property resolver
	 */
	public static PropertyResolver propertyResolver = new DefaultPropertyResolver();

	private static final Pattern HEXA = Pattern.compile("[0-9A-Z]+",
			Pattern.CASE_INSENSITIVE);

	/**
	 * Get a collection has a list. If <code>set</code> is <code>null</code>,
	 * returns an empty list.
	 * 
	 * @param set
	 *            Incoming collection
	 * @param <T>
	 *            Type of items in the set
	 * @return Resulting list
	 */
	public static <T> List<T> asList(Collection<T> set) {
		if (set == null) {
			return Collections.emptyList();
		} else if (set instanceof List<?>) {
			return (List<T>) set;
		} else {
			return new ArrayList<T>(set);
		}
	}

	/**
	 * Gets a property path from a list of properties
	 * 
	 * @param propertyName
	 *            List of properties
	 * @return Property path
	 */
	public static String asPropertyPath(String... propertyName) {
		return StringUtils.join(propertyName, ".");
	}

	/**
	 * Calls a method using introspection.
	 * 
	 * @param target
	 *            Object that holds the method to call
	 * @param methodName
	 *            Method name
	 * @param parameters
	 *            Parameters for the call
	 * @return Result of the call
	 */
	public static Object callMethod(Object target, String methodName,
			Object... parameters) {
		Validate.notNull(target, "Target is required");
		Validate.notEmpty(methodName, "Method name is required");
		try {
			return MethodUtils.invokeMethod(target, methodName, parameters);
		} catch (InvocationTargetException ex) {
			Throwable tex = ex.getTargetException();
			if (tex instanceof RuntimeException) {
				throw (RuntimeException) tex;
			} else {
				throw new LocalizableException(
						StringCodes.STRING_UTILS_METHODCALL_ERROR, tex, target,
						methodName, tex);
			}
		} catch (RuntimeException ex) {
			throw ex;
		} catch (Exception ex) {
			throw new LocalizableException(StringCodes.STRING_UTILS_METHODCALL_ERROR,
					ex, target, methodName, ex);
		}
	}

	/**
	 * Duplicates an object using serialization.
	 * 
	 * @param o
	 *            Object to close
	 * @param <T>
	 *            Type of the object
	 * @return Cloned object
	 * @see SerializationUtils#clone(Serializable)
	 */
	@SuppressWarnings("unchecked")
	public static <T extends Serializable> T clone(T o) {
		return (T) SerializationUtils.clone(o);
	}

	/**
	 * Converts a collection into another using a converter.
	 * 
	 * @param <S>
	 *            Types in the source collection
	 * @param <T>
	 *            Types in the target list
	 * @param sourceCollection
	 *            Source collection
	 * @param targetCollection
	 *            Target collection
	 * @param converter
	 *            Converter
	 */
	public static <S, T> void convert(Collection<? extends S> sourceCollection,
			Collection<T> targetCollection, Function<S, T> converter) {
		Validate.notNull(converter, "Converter is required");
		Validate.notNull(targetCollection, "Target collection is required");
		if (sourceCollection != null) {
			for (S s : sourceCollection) {
				T t = converter.apply(s);
				targetCollection.add(t);
			}
		}
	}

	/**
	 * Converts a collection into another using a converter.
	 * 
	 * @param <S>
	 *            Types in the source collection
	 * @param <T>
	 *            Types in the target list
	 * @param sourceCollection
	 *            Source collection
	 * @param targetCollection
	 *            Target collection
	 * @param converter
	 *            Converter
	 * @deprecated Use {@link #convert(Collection, Collection, Function)}
	 *             instead
	 */
	@Deprecated
	public static <S, T> void convert(Collection<? extends S> sourceCollection,
			Collection<T> targetCollection, ItemConverter<S, T> converter) {
		Validate.notNull(converter, "Converter is required");
		Validate.notNull(targetCollection, "Target collection is required");
		if (sourceCollection != null) {
			for (S s : sourceCollection) {
				T t = converter.convert(s);
				targetCollection.add(t);
			}
		}
	}

	/**
	 * Converts a collection to a list using a function.
	 * 
	 * @param <S>
	 *            Types in the source collection
	 * @param <T>
	 *            Types in the target list
	 * @param sourceCollection
	 *            Source collection
	 * @param function
	 *            Transformation function
	 * @return Target list
	 */
	public static <S, T> List<T> convertToList(
			Collection<? extends S> sourceCollection, Function<S, T> function) {
		Validate.notNull(function, "Converter must not be null");
		if (sourceCollection == null) {
			return null;
		}
		return new ArrayList<T>(Collections2.transform(sourceCollection,
				function));
	}

	/**
	 * Converts a collection to a list using a converter.
	 * 
	 * @param <S>
	 *            Types in the source collection
	 * @param <T>
	 *            Types in the target list
	 * @param sourceCollection
	 *            Source collection
	 * @param converter
	 *            Converter
	 * @return Target list
	 * @deprecated Use
	 *             {@link Lists#transform(List, com.google.common.base.Function)}
	 *             instead.
	 */
	@Deprecated
	public static <S, T> List<T> convertToList(
			Collection<? extends S> sourceCollection,
			ItemConverter<S, T> converter) {
		Validate.notNull(converter, "Converter must not be null");
		if (sourceCollection == null) {
			return null;
		}
		ArrayList<T> targetList = new ArrayList<T>(sourceCollection.size());
		for (S s : sourceCollection) {
			T t = converter.convert(s);
			targetList.add(t);
		}
		return targetList;
	}

	/**
	 * Converts a collection to a list using a {@link PropertyItemConverter}.
	 * 
	 * @param <S>
	 *            Types in the source collection
	 * @param <T>
	 *            Types in the target list
	 * @param sourceCollection
	 *            Source collection
	 * @param property
	 *            Property to extract from each source item
	 * @return Target list
	 */
	public static <S, T> List<T> convertToList(
			Collection<? extends S> sourceCollection, String property) {
		Validate.notEmpty(property, "Property must not be empty");
		if (sourceCollection == null) {
			return null;
		} else {
			return Lists.transform(new ArrayList<S>(sourceCollection),
					new PropertyFunction<S, T>(property));
		}
	}

	/**
	 * Converts a collection to a set using a function.
	 * 
	 * @param <S>
	 *            Types in the source collection
	 * @param <T>
	 *            Types in the target set
	 * @param sourceCollection
	 *            Source collection
	 * @param function
	 *            Transformation function
	 * @return Target set
	 */
	public static <S, T> Set<T> convertToSet(
			Collection<? extends S> sourceCollection, Function<S, T> function) {
		Validate.notNull(function, "function is required");
		if (sourceCollection == null) {
			return null;
		}
		HashSet<T> targetList = new HashSet<T>(sourceCollection.size());
		for (S s : sourceCollection) {
			T t = function.apply(s);
			targetList.add(t);
		}
		return targetList;
	}

	/**
	 * Converts a collection to a set using a converter.
	 * 
	 * @param <S>
	 *            Types in the source collection
	 * @param <T>
	 *            Types in the target set
	 * @param sourceCollection
	 *            Source collection
	 * @param converter
	 *            Converter
	 * @return Target set
	 * @deprecated Use {@link #convertToSet(Collection, Function)} instead
	 */
	@Deprecated
	public static <S, T> Set<T> convertToSet(
			Collection<? extends S> sourceCollection,
			ItemConverter<S, T> converter) {
		Validate.notNull(converter, "Converter is required");
		if (sourceCollection == null) {
			return null;
		}
		HashSet<T> targetList = new HashSet<T>(sourceCollection.size());
		for (S s : sourceCollection) {
			T t = converter.convert(s);
			targetList.add(t);
		}
		return targetList;
	}

	/**
	 * Copy properties from one bean to the other
	 * 
	 * @param destination
	 *            Destination bean
	 * @param origin
	 *            Origin bean
	 */
	public static void copyProperties(Object destination, Object origin) {
		Validate.notNull(destination, "Destination is required");
		Validate.notNull(origin, "Origin is required");
		try {
			BeanUtils.copyProperties(destination, origin);
		} catch (IllegalAccessException e) {
			throw new RuntimeException("Cannot copy properties", e);
		} catch (InvocationTargetException e) {
			throw new RuntimeException("Cannot copy properties",
					e.getTargetException());
		}
	}

	/**
	 * Protected call to <code>Class.forName</code>
	 * 
	 * @param className
	 *            Class name to get the Class instance from
	 * @return Class instance
	 */
	public static Class<?> forClass(String className) {
		Validate.notEmpty(className);
		try {
			Class<?> clazz = Class.forName(className);
			return clazz;
		} catch (Exception ex) {
			throw new LocalizableException(StringCodes.STRING_UTILS_FOR_CLASS_ERROR,
					ex, className);
		}
	}

	/**
	 * Parses a date as ISO 8601
	 * 
	 * @param value
	 *            ISO8601 date
	 * @return Parsed date
	 * @throws ParseException
	 *             If the value cannot be parsed
	 */
	public static Date fromISOFormat(String value) throws ParseException {
		if (StringUtils.isEmpty(value)) {
			return null;
		} else {
			SimpleDateFormat format = new SimpleDateFormat(ISO_DATE_FORMAT);
			return format.parse(value);
		}
	}

	/**
	 * Get an annotation from the class or from its parent classes
	 * 
	 * @param oClass
	 *            Class to search annotation in.
	 * @param aClass
	 *            Annotation to search
	 * @param <A>
	 *            Annotation type
	 * @return Annotation or <code>null</code> if not found.
	 */
	public static <A extends Annotation> A getAnnotation(Class<?> oClass,
			Class<A> aClass) {
		A a = oClass.getAnnotation(aClass);
		if (a != null) {
			return a;
		} else {
			Class<?> superClass = oClass.getSuperclass();
			if (superClass != null) {
				return getAnnotation(superClass, aClass);
			} else {
				return null;
			}
		}
	}

	/**
	 * Gets an annotation for a type.
	 * 
	 * @param <A>
	 *            Type of annotation to look for
	 * @param annotationClass
	 *            Annotation class
	 * @param bean
	 *            Bean that contains the annotation
	 * @return Annotation instance or <code>null</code> if not defined
	 */
	public static <A extends Annotation> A getAnnotation(
			Class<A> annotationClass, Object bean) {
		Class<? extends Object> beanClass = bean.getClass();
		return beanClass.getAnnotation(annotationClass);
	}

	/**
	 * Gets the value for a constant.
	 * 
	 * @param type
	 *            Type that contains the constant
	 * @param name
	 *            Name of the constant
	 * @return Constant value
	 */
	public static Object getConstant(Class<?> type, String name) {
		try {
			Field constant = type.getField(name);
			return constant.get(null);
		} catch (Exception ex) {
			throw new RuntimeException("Cannot get constant " + name + " from "
					+ type.getName(), ex);
		}
	}

	/**
	 * List of properties for a set of objects
	 * 
	 * @param items
	 *            List of objects to loop on
	 * @param property
	 *            Property to get for each object
	 * @return List of all properties values
	 */
	public static List<Object> getProperties(Collection<?> items,
			String property) {
		ArrayList<Object> properties = new ArrayList<Object>();
		for (Iterator<?> i = items.iterator(); i.hasNext();) {
			Object item = i.next();
			try {
				Object value = PropertyUtils.getProperty(item, property);
				properties.add(value);
			} catch (Exception ex) {
				log.error("Cannot get property " + property + " on " + item);
			}
		}
		// Ok
		return properties;
	}

	/**
	 * Gets a property from a bean
	 * 
	 * @param bean
	 *            Bean
	 * @param property
	 *            Property name
	 * @return Property value
	 */
	public static Object getProperty(Object bean, String property) {
		return propertyResolver.getProperty(bean, property);
	}

	/**
	 * Gets an annotation for a property. The annotation may be associated with
	 * either the getter or the setter method of the property.
	 * 
	 * @param <A>
	 *            Type of annotation to look for
	 * @param annotationClass
	 *            Annotation class
	 * @param bean
	 *            Bean that contains the property
	 * @param propertyName
	 *            Property name
	 * @return Annotation instance or <code>null</code> if not defined
	 */
	public static <A extends Annotation> A getPropertyAnnotation(
			Class<A> annotationClass, Object bean, String propertyName) {
		if (bean == null) {
			throw new IllegalArgumentException("bean cannot be null");
		}
		try {
			// Gets the property descriptor
			PropertyDescriptor propertyDescriptor = PropertyUtils
					.getPropertyDescriptor(bean, propertyName);
			// Gets the write method
			Method writeMethod = propertyDescriptor.getWriteMethod();
			if (writeMethod != null) {
				A annotation = writeMethod.getAnnotation(annotationClass);
				if (annotation != null) {
					return annotation;
				}
			}
			// Gets the read method
			Method readMethod = propertyDescriptor.getReadMethod();
			if (readMethod != null) {
				A annotation = readMethod.getAnnotation(annotationClass);
				if (annotation != null) {
					return annotation;
				}
			}
			// Not found
			return null;
		} catch (IllegalAccessException e) {
			throw new LocalizableException(
					StringCodes.UTILS_PROPERTY_ANNOTATION_CANNOT_ACCESS, e,
					annotationClass.getName(), bean, propertyName);
		} catch (InvocationTargetException e) {
			throw new LocalizableException(
					StringCodes.UTILS_PROPERTY_ANNOTATION_CANNOT_ACCESS, e,
					annotationClass.getName(), bean, propertyName);
		} catch (NoSuchMethodException e) {
			throw new LocalizableException(
					StringCodes.UTILS_PROPERTY_ANNOTATION_CANNOT_ACCESS, e,
					annotationClass.getName(), bean, propertyName);
		}
	}

	public static PropertyResolver getPropertyResolver() {
		return propertyResolver;
	}

	/**
	 * Gets a resource URL from a reference class and a resource path. The
	 * resource URL is determined by computing the resource path from the
	 * reference class.
	 * 
	 * If the reference class is <code>null</code>, the path is computed using
	 * the current class (i.e. the {@link Utils} class).
	 * 
	 * If the URL cannot be accessed, a LocalizableException with code
	 * {@link StringCodes#STRING_ERROR_RESOURCE_NOT_FOUND} is thrown.
	 * 
	 * @param referenceClass
	 *            Reference for the resource path
	 * @param path
	 *            Resource path from the reference class
	 * @return Resource URL
	 * @throws LocalizableException
	 *             If the URL cannot be found
	 */
	public static URL getResource(Class<?> referenceClass, String path)
			throws LocalizableException {
		URL url;
		if (referenceClass != null) {
			url = referenceClass.getResource(path);
		} else {
			url = Utils.class.getResource(path);
		}
		if (url != null) {
			return url;
		} else {
			throw new LocalizableException(
					StringCodes.STRING_ERROR_RESOURCE_NOT_FOUND, path);
		}
	}

	/**
	 * Creates a new instance
	 * 
	 * @param clazz
	 *            Type of the instance to create
	 * @param <T>
	 *            Type of the class
	 * @return Instance of the class
	 */
	public static <T> T newInstance(Class<T> clazz) {
		try {
			T o = clazz.newInstance();
			return o;
		} catch (LocalizableException ex) {
			throw ex;
		} catch (Exception ex) {
			throw new LocalizableException(
					StringCodes.STRING_UTILS_NEW_INSTANCE_ERROR, ex,
					clazz.getName());
		}
	}

	/**
	 * Creates a new instance with a unique parameter in the constructor
	 * 
	 * @param clazz
	 *            Type of the instance to create
	 * @param <T>
	 *            Type of the class
	 * @param parameter
	 *            Parameter for the constructor
	 * @return Instance of the class
	 */
	public static <T> T newInstance(Class<T> clazz, Object parameter) {
		try {
			if (parameter == null) {
				Constructor<?>[] constructors = clazz.getConstructors();
				for (Constructor<?> constructor : constructors) {
					Class<?>[] parameterTypes = constructor.getParameterTypes();
					if (parameterTypes.length == 1) {
						// Takes the first constructor
						@SuppressWarnings("unchecked")
						T o = (T) constructor
								.newInstance(new Object[] { null });
						return o;
					}
				}
				// No constructor has been found
				throw new RuntimeException(
						"Cannot find any constructor with one parameter for "
								+ clazz.getName());
			} else {
				Constructor<T> constructor = clazz.getConstructor(parameter
						.getClass());
				T o = constructor.newInstance(parameter);
				return o;
			}
		} catch (LocalizableException ex) {
			throw ex;
		} catch (Exception ex) {
			throw new LocalizableException(
					StringCodes.STRING_UTILS_NEW_INSTANCE_ERROR, ex,
					clazz.getName());
		}
	}

	/**
	 * Creates a new instance of a class given its name.
	 * 
	 * @param className
	 *            Type name of the instance to create
	 * @return Instance of the given type
	 */
	public static Object newInstance(String className) {
		try {
			Class<?> clazz = Class.forName(className);
			Object o = clazz.newInstance();
			return o;
		} catch (LocalizableException ex) {
			throw ex;
		} catch (Exception ex) {
			throw new LocalizableException(
					StringCodes.STRING_UTILS_NEW_INSTANCE_ERROR, ex, className,
					ex);
		}
	}

	/**
	 * Creates a new instance of a class given its name and a set of properties
	 * 
	 * @param className
	 *            Type name of the instance to create
	 * @param params
	 *            Parameters to set in the newly created instance
	 * @return Instance of the given type where properties have been set
	 *         according to the given parameters
	 * @see #setProperties(Object, ParameterSet)
	 * @see #newInstance(String)
	 */
	public static Object newInstance(String className, ParameterSet params) {
		Object o = newInstance(className);
		setProperties(o, params);
		return o;
	}

	/**
	 * Parses a query
	 * 
	 * @param query
	 *            Query to parse
	 * @return Extracted parameters. The returned map can be empty but never
	 *         <code>null</code>.
	 */
	public static Map<String, String> parseQuery(String query) {
		HashMap<String, String> params = new HashMap<String, String>();
		if (StringUtils.isNotBlank(query)) {
			String[] tokens = StringUtils.split(query, "&");
			for (String token : tokens) {
				String[] pair = StringUtils.split(token, "=");
				if (pair.length == 1) {
					params.put(pair[0], pair[0]);
				} else if (pair.length == 2) {
					params.put(pair[0], pair[1]);
				}
			}
		}
		// Ok
		return params;
	}

	/**
	 * Parses a URL query
	 * 
	 * @param url
	 *            URL to parse
	 * @return Extracted parameters. The returned map can be empty but never
	 *         <code>null</code>.
	 */
	public static Map<String, String> parseQuery(URL url) {
		String query = url.getQuery();
		return parseQuery(query);
	}

	/**
	 * Reads a text resource from a reference class and a resource path. The
	 * resource URL is determined by computing the resource path from the
	 * reference class.
	 * 
	 * If the reference class is <code>null</code>, the path is computed using
	 * the current class (i.e. the {@link IOUtils} class).
	 * 
	 * If the resource cannot be accessed, a LocalizableException with code
	 * {@link StringCodes#STRING_ERROR_RESOURCE_NOT_FOUND} is thrown.
	 * 
	 * @param referenceClass
	 *            Reference for the resource path
	 * @param path
	 *            Resource path from the reference class
	 * @param encoding
	 *            Encoding of the target resource
	 * @return Resource content
	 * @throws LocalizableException
	 *             If the resource cannot be found
	 */
	public static String readTextResource(Class<?> referenceClass, String path,
			String encoding) throws LocalizableException {
		InputStream in;
		if (referenceClass != null) {
			in = referenceClass.getResourceAsStream(path);
		} else {
			in = IOUtils.class.getResourceAsStream(path);
		}
		if (in != null) {
			try {
				BufferedReader reader = new BufferedReader(
						new InputStreamReader(in, encoding));
				try {
					StringBuilder buffer = new StringBuilder();
					char[] t = new char[1024];
					int read;
					while ((read = reader.read(t)) > 0) {
						buffer.append(t, 0, read);
					}
					return buffer.toString();
				} finally {
					reader.close();
				}
			} catch (IOException ex) {
				throw new LocalizableException("Utils.ResourceIOError", ex, path, ex);
			}
		} else {
			throw new LocalizableException("Utils.ResourceNotFound", path);
		}
	}

	/**
	 * Set the properties of an object
	 * 
	 * @param o
	 *            Object to set
	 * @param params
	 *            Parameters to set in the given object
	 */
	public static void setProperties(Object o, ParameterSet params) {
		if (params != null) {
			for (Map.Entry<String, String> entry : params.getParams()
					.entrySet()) {
				String name = entry.getKey();
				String value = entry.getValue();
				try {
					BeanUtils.setProperty(o, name, value);
				} catch (Exception ex) {
					throw new LocalizableException(
							StringCodes.STRING_UTILS_SET_PROPERTIES_ERROR, ex,
							o, name, value, ex.getLocalizedMessage());
				}
			}
		}
	}

	/**
	 * Sets a property on a bean
	 * 
	 * @param bean
	 *            Bean
	 * @param property
	 *            Property name
	 * @param value
	 *            Property value
	 */
	public static void setProperty(final Object bean, final String property,
			final Object value) {
		if (bean instanceof Component) {
			if (SwingUtilities.isEventDispatchThread()) {
				propertyResolver.setProperty(bean, property, value);
			} else {
				SwingUtilities.invokeLater(new Runnable() {

					@Override
					public void run() {
						try {
							propertyResolver.setProperty(bean, property, value);
						} catch (Exception e) {
							log.error(
									String.format(
											"Cannot set property %s on bean %s with value %s",
											property, bean, value), e);
						}
					}
				});
			}
		} else {
			propertyResolver.setProperty(bean, property, value);
		}
	}

	public static void setPropertyResolver(PropertyResolver propertyResolver) {
		Utils.propertyResolver = propertyResolver;
	}

	public static String simpleToString(Object o, String nullString) {
		String value = ObjectUtils.toString(o, nullString);
		// Default toString?
		String num = StringUtils.substringAfterLast(value, "@");
		if (HEXA.matcher(num).matches()) {
			// if (REGEX_DEFAULT_TO_STRING.matcher(value).matches()) {
			String beforeA = StringUtils.substringBeforeLast(value, "@");
			return StringUtils.substringAfterLast(beforeA, ".");
		} else {
			return value;
		}
	}

	/**
	 * Formats a date as ISO 8601 format
	 * 
	 * @param date
	 *            Date to format
	 * @return Formatted date
	 */
	public static String toISOFormat(Date date) {
		if (date == null) {
			return null;
		} else {
			SimpleDateFormat format = new SimpleDateFormat(ISO_DATE_FORMAT);
			return format.format(date);
		}
	}

	/**
	 * Converts a sequence of strings to a map. Each string must have the format
	 * : &lt;key&gt;=&lt;value&gt;.
	 * 
	 * @param strings
	 *            List of strings
	 * @return Corresponding map (not <code>null</code>)
	 */
	public static Map<String, String> toMap(String[] strings) {
		HashMap<String, String> map = new HashMap<String, String>();
		if (strings != null && strings.length != 0) {
			for (String token : strings) {
				if (StringUtils.isNotBlank(token)) {
					String[] tokens = StringUtils.split(token, "=");
					if (tokens.length == 1) {
						map.put(tokens[0], tokens[0]);
					} else {
						map.put(tokens[0], tokens[1]);
					}
				}
			}
		}
		return map;
	}

	private Utils() {
	}
}
