/*
 * Created on 8 oct. 2005
 */
package net.sf.doolin.util;

import org.apache.commons.lang.ArrayUtils;
import org.apache.commons.lang.StringUtils;
import org.apache.commons.lang.Validate;

/**
 * Definition of a version on one or more numbers.
 * 
 * @author Damien Coraboeuf
 */
public class Version implements Comparable<Version> {

	/**
	 * Position of the major version number
	 */
	public static final int MAJOR = 0;

	/**
	 * Position of the minor version number
	 */
	public static final int MINOR = 1;

	/**
	 * Position of the patch version number
	 */
	public static final int PATCH = 2;

	private final int[] versions;

	/**
	 * Constructor
	 * 
	 * @param numbers
	 *            Version numbers, from the most major to the most minor
	 */
	public Version(int... numbers) {
		Validate.notNull(numbers, "Numbers is required");
		Validate.isTrue(numbers.length > 0,
				"At least one version number must be specified");
		this.versions = ArrayUtils.clone(numbers);
	}

	/**
	 * Parsing of a version number
	 * 
	 * @param value
	 *            String to parse.
	 */
	public Version(String value) {
		String[] tokens = StringUtils.split(value, ".");
		int count = tokens.length;
		if (count < 1) {
			throw new IllegalArgumentException("Not enough digits in version "
					+ value);
		} else {
			this.versions = new int[count];
			for (int i = 0; i < count; i++) {
				int number = Integer.parseInt(tokens[i], 10);
				this.versions[i] = number;
			}
		}
	}

	/**
	 * 0.0.0 version
	 */
	public Version() {
		this(0, 0, 0);
	}

	/**
	 * @param position
	 *            Position to get (must be &gt;= 0)
	 * @return Returns the version number at the specified position. If this
	 *         position is not defined, returns 0.
	 */
	public int getNumber(int position) {
		if (position < 0) {
			throw new IllegalArgumentException("Position must be >= 0");
		} else if (position < this.versions.length) {
			return this.versions[position];
		} else {
			return 0;
		}
	}

	/**
	 * @return Returns the major.
	 */
	public int getMajor() {
		return getNumber(MAJOR);
	}

	/**
	 * @return Returns the minor.
	 */
	public int getMinor() {
		return getNumber(MINOR);
	}

	/**
	 * @return Returns the patch.
	 */
	public int getPatch() {
		return getNumber(PATCH);
	}

	/**
	 * @see java.lang.Object#hashCode()
	 */
	@Override
	public int hashCode() {
		int total = 0;
		for (int i = 0; i < this.versions.length; i++) {
			total += total * 10 + this.versions[i];
		}
		return total;
	}

	/**
	 * Returns the number of versions numbers
	 * 
	 * @return Number of version numbers
	 */
	public int getDimension() {
		return this.versions.length;
	}

	/**
	 * @see java.lang.Object#equals(java.lang.Object)
	 */
	@Override
	public boolean equals(Object obj) {
		if (obj instanceof Version) {
			Version v = (Version) obj;
			int max = Math.max(getDimension(), v.getDimension());
			for (int i = 0; i < max; i++) {
				int a = getNumber(i);
				int b = v.getNumber(i);
				if (a != b) {
					return false;
				}
			}
			return true;
		} else {
			return false;
		}
	}

	/**
	 * @see java.lang.Object#toString()
	 */
	@Override
	public String toString() {
		return StringUtils.join(ArrayUtils.toObject(this.versions), ".");
	}

	/**
	 * Gets a copy of the version array.
	 * 
	 * @return Version numbers
	 */
	public int[] toArray() {
		return ArrayUtils.clone(this.versions);
	}

	/**
	 * Compares two versions
	 * 
	 * @param v
	 *            Version to compare to.
	 * @return Result of comparison.
	 */
	public int compareTo(Version v) {
		int max = Math.max(getDimension(), v.getDimension());
		for (int i = 0; i < max; i++) {
			int a = getNumber(i);
			int b = v.getNumber(i);
			if (a != b) {
				return a - b;
			}
		}
		return 0;
	}
}
