package net.sf.doolin.util.unit;

import java.math.BigDecimal;

import org.apache.commons.lang.ObjectUtils;
import org.apache.commons.lang.StringUtils;
import org.apache.commons.lang.builder.HashCodeBuilder;

/**
 * Numeric value associated with a unit.
 * 
 * @author Damien Coraboeuf
 */
public class ValueUnit {

	/**
	 * Converts a formatted string to a numeric value and a unit.
	 * 
	 * @param string
	 *            Formatted string.
	 * @return Value + unit or <code>null</code> if the string is empty
	 * @throws IllegalArgumentException
	 *             If the provided string cannot be parsed
	 */
	public static ValueUnit valueOf(String string) {
		if (StringUtils.isBlank(string)) {
			return null;
		} else {
			Unit[] units = Unit.values();
			for (Unit unit : units) {
				String symbol = unit.getSymbol();
				if (string.endsWith(symbol)) {
					String numericValue = StringUtils.substringBefore(string,
							symbol);
					try {
						Number number = new BigDecimal(numericValue);
						return new ValueUnit(number, unit);
					} catch (NumberFormatException ex) {
						throw new IllegalArgumentException("Cannot parse "
								+ string, ex);
					}
				}
			}
			throw new IllegalArgumentException("Cannot parse " + string);
		}
	}

	private Number value;

	private Unit unit;

	/**
	 * Constructor with no predefined value or unit.
	 */
	public ValueUnit() {
	}

	/**
	 * Constructor with a predefined value and unit.
	 * 
	 * @param value
	 *            Numeric value
	 * @param unit
	 *            Unit
	 */
	public ValueUnit(Number value, Unit unit) {
		this.value = value;
		this.unit = unit;
	}

	/**
	 * Converts this value to another unit.
	 * 
	 * @param targetUnit
	 *            Target unit
	 * @param reference
	 *            Reference value for the percentage unit
	 * @return New value
	 */
	public ValueUnit convert(Unit targetUnit, Number reference) {
		// No need of conversion for same units
		if (this.unit == targetUnit) {
			return new ValueUnit(this.value, this.unit);
		} else if (this.unit == Unit.PERCENT) {
			/*
			 * Special case
			 * 
			 * The new value is a percentage of the reference
			 */
			double newValue = this.value.doubleValue()
					* reference.doubleValue() / 100;
			return new ValueUnit(newValue, targetUnit);
		} else {
			// Converts current value to pixels
			double pixels = this.value.doubleValue()
					* this.unit.getPixels(reference);
			// Converts pixels to new unit
			Number newValue = pixels / targetUnit.getPixels(reference);
			// Ok
			return new ValueUnit(newValue, targetUnit);
		}
	}

	@Override
	public boolean equals(Object obj) {
		if (obj instanceof ValueUnit) {
			ValueUnit vu = (ValueUnit) obj;
			return ObjectUtils.equals(this.unit, vu.unit)
					&& ObjectUtils.equals(this.value, vu.value);
		} else {
			return false;
		}
	}

	/**
	 * @return Associated unit
	 */
	public Unit getUnit() {
		return this.unit;
	}

	/**
	 * @return Associated value
	 */
	public Number getValue() {
		return this.value;
	}

	@Override
	public int hashCode() {
		return new HashCodeBuilder().append(this.value).append(this.unit)
				.hashCode();
	}

	/**
	 * @param unit
	 *            Associated unit
	 */
	public void setUnit(Unit unit) {
		this.unit = unit;
	}

	/**
	 * @param value
	 *            Associated value
	 */
	public void setValue(Number value) {
		this.value = value;
	}

	@Override
	public String toString() {
		return this.value + this.unit.getSymbol();
	}

}
