package net.sf.doolin.util.xml;

import java.util.ArrayList;
import java.util.List;

import org.slf4j.Logger;
import org.xml.sax.ErrorHandler;
import org.xml.sax.SAXException;
import org.xml.sax.SAXParseException;

/**
 * This class implements a simple error handler that just stores the warnings
 * and errors that occur during an XML parsing.
 * 
 * @author Damien Coraboeuf
 */
public class SimpleErrorHandler implements ErrorHandler {

	private ArrayList<SAXParseException> warnings = new ArrayList<SAXParseException>();

	private ArrayList<SAXParseException> errors = new ArrayList<SAXParseException>();

	public void error(SAXParseException exception) throws SAXException {
		this.errors.add(exception);
	}

	public void fatalError(SAXParseException exception) throws SAXException {
		this.errors.add(exception);
	}

	/**
	 * Returns the list of errors
	 * 
	 * @return List of errors
	 */
	public List<SAXParseException> getErrors() {
		return this.errors;
	}

	/**
	 * Returns the list of warnings
	 * 
	 * @return List of warnings
	 */
	public List<SAXParseException> getWarnings() {
		return this.warnings;
	}

	/**
	 * Returns if the handler contains some errors.
	 * 
	 * @return <code>true</code> if some errors have been generated.
	 */
	public boolean hasErrors() {
		return !this.errors.isEmpty();
	}

	/**
	 * Returns if the handler contains some warnings.
	 * 
	 * @return <code>true</code> if some warnings have been generated.
	 */
	public boolean hasWarnings() {
		return !this.warnings.isEmpty();
	}

	/**
	 * Logs all the warnings
	 * 
	 * @param log
	 *            Logger to use
	 */
	public void logWarnings(Logger log) {
		for (SAXParseException ex : this.warnings) {
			log.warn("Parsing warning : " + ex.getMessage());
		}
	}

	/**
	 * Throw the first error is any.
	 * 
	 * @throws SAXParseException
	 *             Thrown exception
	 */
	public void throwAnyError() throws SAXParseException {
		if (hasErrors()) {
			throw this.errors.get(0);
		}
	}

	/**
	 * Throw the first error or warning is any.
	 * 
	 * @throws SAXParseException
	 *             Thrown exception
	 */
	public void throwAnyWarningOrError() throws SAXParseException {
		throwAnyError();
		if (hasWarnings()) {
			throw this.warnings.get(0);
		}
	}

	public void warning(SAXParseException exception) throws SAXException {
		this.warnings.add(exception);
	}
}
